﻿#pragma TextEncoding="UTF-8"
#pragma rtGlobals=3
#pragma IgorVersion=7

// feedback? ideas? send me a note: https: www.wavemetrics.com/user/tony

// This file should be saved in the Igor Procedures folder (look in the 
// Help menu for "Show Igor Pro User Files"). 

// If you want Igor to automatically reload updated files, choose 
// Misc->Miscellaneous Settings, click Text Editing in the left pane, 
// click the External Editor tab, and select Reload Automatically, As 
// Soon as Modification is Detected. 

// If you're an end-user of packages that can be updated this way, that's
// all you need to do.

// ------------------------------------------------------------------------

// If you're developing a package and wish to use this updater: 

// 1. Copy and paste this block of pragma statements and constants to the procedure file you 
// want to keep updated. Edit values to match your package - the values here are for this file.

#pragma version=2.38 // Edit version number - must be same as release version on IgorExchange
#pragma moduleName=updater // Edit regular module name
static constant kCHECK_FOR_UPDATES=1  // set to 1 to check for updates on init
static strconstant ksSHORTNAME="Updater" // Must correspond to shortname of IgorExchange release
									// (ksSHORTNAME isn't needed if moduleName matches project shortname)
 
// 2. Insert this code in your procedure file (without //):

//static function CheckForUpdates()
//	if(kCHECK_FOR_UPDATES)
//		#if (exists("Updater#UpdateCheck")==6)
//			return Updater#UpdateCheck("fileloc:"+FunctionPath(""))
//		#endif
//	endif
//	return 0
//end

// 3. Add this to your package initialization function:

//	if (CheckForUpdates())
//		return 0 // this file has been updated; quit to allow user to reload
//	endif

// ---------------------------------------------------------------------------

// -- Details
// If the project release is a zip archive, Updater downloads the archive
// to a temporary location on the desktop, and moves all of the contents 
// to a location in the User Procedures folder based on the location of 
// the ipf file to be updated. For this to work, the zip archive had 
// better have a straightforward structure. Caveat emptor!

// 	-- Wishlist
// Would like to know current status of Misc settings for reloading 
// updated files.
// Provide a way to rollback to earlier releases, or give a list of 
// releases for the user to choose from.

// -----------------------------------------------------------------------------

// -- code for use in independent modules:

//static function CheckForUpdates()
//	
//	if (kCHECK_FOR_UPDATES==0)
//		return 0
//	endif
//	
//	string keyStr=""
//	keyStr=ReplaceStringByKey("fileloc", keyStr, FunctionPath(""))
//	
//// find out whether Updater is installed
//	#if (exists("Updater#UpdateCheck")==6) // this will be 0 if running as an independent module	
//		return Updater#UpdateCheck(keyStr)
//	#endif
//
//	if(stringmatch(GetIndependentModuleName(), "ProcGlobal"))
//		return 0
//	endif
//	
//// Running as independent module, so will have to use execute.	
//// Find out whether procedures are compiled:
//
//	variable /G V_updated=0
//	variable isUpdated=0
//	string cmd, infoStr=FunctionInfo("ProcGlobal#updater#UpdateCheck", "Updater.ipf")
//
//	if (stringmatch(StringByKey("MODULE", infoStr), "updater"))	// updater is available AND compiled
//		sprintf cmd, "V_updated=ProcGlobal#Updater#UpdateCheck(\"%s\")", keyStr
//		execute cmd
//		isUpdated=V_updated
//	endif
//	
//	killvariables /Z V_updated
//
//	return isUpdated 
//end

// ------------------------------------------------------------------------------------------------------------

static constant kFREQUENCY=604800 // do not check more frequently than this. weekly=604800.
// you can override this setting by adding a copy of the above line to your procedure file and editing the value
// you can also override the default location for updates by adding something like this to your procedure:
// static strconstant ksLOCATION="https://www.wavemetrics.com/project-releases/1234" 

// ------------------------------------------------------------------------------------------------------------

static constant kPageTimeout=2 // time limit for page downloads
static constant kFileTimeout=10 // time limit for file downloads
static strconstant ksIgnoreFilesList=".DS_Store;" // list of files that shouldn't be copied, wildcards okay
static strconstant ksIgnoreFoldersList="__MACOSX;" // list of folders that shouldn't be copied, wildcards okay

menu "Help"
	"Tony's Updater: Check for new releases", /Q, updater#UpdateAll()
end

// We'll keep a very small binary file in the updater preferences folder to
// store the last update check time for each project that works with updater
static structure LastUpdateStructure
	uint32	 version		// Preferences structure version number. 100 means 1.00.
	uint32 lastUpdate	// 32 bits should be good for another 20 years or so...
	uint32 reserved[8]	// Reserved for future use
endStructure


// Wrapper function for update()
// This function is called from the file to be checked
static function UpdateCheck(keyList)
	string keyList
	
	variable updated=0
	
	NVAR /Z v_updater=v_updater // if this exists we want to check all open procedures for possible updates
	if (!NVAR_Exists(v_updater)) // avoid repeatedly checking this file
		// first find out whether this file needs updating
		string keyListThisFile=""
		keyListThisFile=ReplaceStringByKey("fileloc", keyListThisFile, FunctionPath(""))
		updated+=update(keyListThisFile)
	endif
	// now check the calling procedure file
	updated+=update(keyList)

	return updated
end


// Looks for procedure windows with defined modules that contain the 
// static CheckForUpdates function, and executes that function wherever 
// it is found. Only works for procedure files in which a regular module 
// name is defined.
static function UpdateAll()
	
	// Use a global variable to let UpdateCheck() function in this 
	// procedure know that we want to ignore the frequency setting and 
	// check now for updates to all files that use updater and have 
	// kCheckForUpdates set.
	variable /g v_updater=0
	
	string procList="", IMlist="", moduleList="", moduleStr="", IMstr="", cmd=""
	variable i,j, imax, jmax
	
	execute "SetIgorOption IndependentModuleDev=1" // would be good to record current status before setting this
	procList=winlist("*",";","WIN:128")
	execute "SetIgorOption IndependentModuleDev=0"
	IMlist=IndependentModuleList(";")	
	
	// loop through procedure windows
	imax=itemsinlist(procList)
	for(i=0; i<imax; i+=1)
		IMstr=""
		moduleStr=FunctionInfo("CheckForUpdates", stringfromlist(i,procList))
		moduleStr=StringByKey("MODULE", moduleStr)
		if(strlen(moduleStr)==0) // didn't find CheckForUpdates function
			// check to see if it's an independent module
			jmax=itemsinlist(IMlist)
			for(j=0;j<jmax;j+=1)
				IMstr=stringfromlist(j,IMlist)
				moduleStr=FunctionInfo(IMstr+"#CheckForUpdates", stringfromlist(i,procList))
				IMstr=StringByKey("INDEPENDENTMODULE", moduleStr)
				moduleStr=StringByKey("MODULE", moduleStr)
				if (strlen(IMstr))
					break
				endif
				IMstr=""
			endfor
		endif
		
		if (strlen(moduleStr))
			moduleStr+="#"
		endif
		if (strlen(IMstr))
			IMstr+="#"
		endif
		
		if (strlen(moduleStr)|| strlen(IMstr))
			moduleList=addlistitem(IMstr+moduleStr,moduleList)
		endif
	endfor

	// remove this procedure from the list
	moduleList=removefromlist("updater#",moduleList)
	
	// check this file first
	update("fileloc:"+FunctionPath(""))
	
	// now check other files
	imax=itemsinlist(moduleList)
	for(i=0; i<imax; i+=1)
		// make sure that we can execute the function within target procedure window
		v_updater=0
		sprintf cmd "v_updater=(exists(\"%sCheckForUpdates\")==6)", stringfromlist(i,moduleList)
		execute cmd
		if(v_updater)
			sprintf cmd "%sCheckForUpdates()", stringfromlist(i,moduleList)
			execute cmd
		endif
	endfor
	
	killvariables /Z v_updater
	return 1	
end

// Minimum amount of information required in file to be updated by this function:
// # pragma version
// static strconstant ksSHORTNAME (or #pragma moduleName corresponding to IgorExchange project shortname).
static function update(keyList)	
	string keyList
	
	// for now, these settings revert to default values when this file is updated, 
	// so don't rely on them sticking if you change them.
	
	// choose default download path here (downloads folder will be created within existing folder):
	string downloadPathStr=SpecialDirPath("Desktop",0,0,0)+"ACW_PackageInstallerTemp:"
	// choose where to save copies of any files that are replaced (set to "" for no backup)
	string backupPathStr=SpecialDirPath("Desktop",0,0,0) // this location must exist, a subfolder may be created
	//string backupPathStr=""
	
	variable testing=0, chatty=0 // set to 1 for a lot of dialog 
	variable autoloadOption=0 // 1 will prompt user to create Igor Procedures alias
	
	variable releaseVersion, releaseIgorVersion, checkFrequency, err, isZip, isWin
	variable localVersion=0, selStart=-1, selEnd=-1
	variable currentIgorVersion=str2num(StringByKey("IGORFILEVERSION", igorinfo(3)))
	variable refNum=0
	variable flagI, i
	
	string procPathStr="", shortName="", url="", cmd=""
	string prefsFile="", projectName="", system="", fileType=""
	string packagePathStr="", archivePathStr=""
	string destPathStr=""
	string archiveName="", ipfName="", fileName=""
	string overwriteList="", fileList="", folderList=""
	
	system=selectString(stringmatch(StringByKey("OS", igorinfo(3))[0,2],"Mac"), "Windows", "Macintosh")
	isWin=stringmatch(system[0,2],"Win")
	procPathStr=stringByKey("fileloc", keyList) // path to procedure file to be updated
	if (strlen(procPathStr)==0)
		return 0
	endif
	ipfName=parseFilePath(0, procPathStr, ":", 1, 0) // name of procedure file to be updated
	packagePathStr=parseFilePath(1, procPathStr, ":", 1, 0) // path to folder containing procedure file
	
	// how frequently to run check
	Grep /Q/E="(?i)^[\s]*static[\s]*constant[\s]*kFREQUENCY[\s]*=" /LIST/Z procPathStr
	s_value=lowerStr(trimString(s_value, 1))
	sscanf s_value, "static constant kfrequency = %d", checkFrequency
	if(V_flag==0)
		checkFrequency=kFREQUENCY // default to value in THIS file
	endif
	
	NVAR /Z v_updater=v_updater // if this exists we want to check all open procedures for possible updates
	if (NVAR_Exists(v_updater))
		checkFrequency=0 // check now regardless of frequency setting 
	endif
	
	// find project shortname
	shortName=getUpdaterStringConst("ksSHORTNAME", procPathStr)
	if (strlen(shortName)==0)
		Grep /Q/E="(?i)#pragma[\s]*moduleName[\s]*=" /LIST/Z procPathStr
		if (strlen(s_value)==0)
			Grep /Q/E="(?i)#pragma[\s]*IndependentModule[\s]*=" /LIST/Z procPathStr
			if (strlen(s_value)==0)
				printf "Updater: No module name defined in %s\r" procPathStr
				return 0
			endif
		endif
		s_value=removeending(s_value,";")
		selStart=strsearch(s_value, "=", 0)+1
		selEnd=strsearch(s_value, "//", 0)
		selEnd = selEnd==-1 ? strlen(s_value)-1 : selEnd-1
		shortName=trimstring(s_value[selStart,selEnd])
	endif
	
	prefsFile=shortName+"LastUpdate.bin"
	
	// check for time of last update check
	STRUCT LastUpdateStructure s
	LoadPackagePreferences "updater", prefsFile, 1, s
	if (V_flag!=0 || V_bytesRead==0)
		s.version=200
		s.lastUpdate=0
		SavePackagePreferences "updater", prefsFile, 1, s		
	endif
	
	if(testing)
		print "\rPackage preferences structure:"
		print s
	endif
	
	if ((datetime-s.lastUpdate)<checkFrequency && !testing)
		return 0 // silently give up
	endif	
	
	// where do we look for updates?
	// first check whether specified in file
	url=getUpdaterStringConst("ksLOCATION", procPathStr)
	if (strlen(url)==0)
		// check the project page on IgorExchange
		sprintf url, "https://www.wavemetrics.com/project/%s", shortName
			
		URLRequest /TIME=(kPageTimeout)/Z url=url
		if(V_flag)
			printf "Installer could not load %s\r", url
			return 0
		endif
		// web page text should be in S_serverResponse

		// find the link to the 'all releases' page for this project
		selEnd=strsearch(S_serverResponse, ">View all releases</a>", 0, 2)
		selStart=strsearch(S_serverResponse, "href=", selEnd, 3)
		if (selEnd==-1 || selStart==-1)
			return 0
		endif
		url="https://www.wavemetrics.com"+S_serverResponse[selStart+6,selEnd-2]
	endif
	
	// version number of local file	
	Grep /Q/E="(?i)^#pragma[\s]*version[\s]*=" /LIST/Z procPathStr
	s_value=lowerStr(trimString(s_value, 1))
	sscanf s_value, "#pragma version = %f", localVersion
	if (V_flag!=1 || localVersion<=0)
		printf "Could not find version pragma in %s\r", ipfName
		return 0
	endif
	
	if(chatty)
		sprintf cmd, "Check for updates to %s?", shortName
		doalert 1, cmd
		if (v_flag==2)
			return 0
		endif
	endif
	printf "%s: Checking for updates... ", shortName
	
	s.lastUpdate=datetime // will be rounded to 32 bit integer
	SavePackagePreferences "updater", prefsFile, 1, s
	
	// download the "all releases" page	
	URLRequest /TIME=(kPageTimeout)/Z url=url
	if(V_flag)
		sprintf cmd, "Could not load %s", url
		print cmd
		if(chatty)
			doalert 0, cmd
		endif
		return 0
	endif
	
	// Look for new releases. If the web page location is not IgorExchange, 
	// ParseReleases will have to be changed to take into account the format 
	// of the web page.	
	wave /T w_releases=ParseReleases(S_serverResponse)
	if (dimsize(w_releases,0)==0)
		sprintf cmd, "Could not find packages at %s.\r", url
		cmd+="\tPlease check for new versions at https://www.wavemetrics.com/projects"
		if (chatty)
			doalert 0, cmd
		endif
		print cmd
		return 0
	endif	
	
	url=""
	for(i=0;i<dimsize(w_releases, 0);i+=1)
		projectName = w_releases[i][0]
		sscanf (w_releases[i][1])[5,inf], "%f.x-%f", releaseIgorVersion, releaseVersion
		
		if(i==0 && localVersion>=releaseVersion)
			sprintf cmd, "Current version (%g) is the most recent release\r", localVersion
			if (chatty)
				doalert 0, cmd
			endif
			print cmd
			return 0
		endif
		
		if(releaseVersion>localVersion)
			if(currentIgorVersion<releaseIgorVersion)
				printf "%s %g requires Igor Pro version >= %g\r", projectName, releaseVersion, releaseIgorVersion
			elseif(FindListItem(system, w_releases[i][5])==-1)
				printf "%s %g not available for %s\r", projectName, releaseVersion, system
			else
				url=w_releases[i][4]
				break
			endif
		endif	
	endfor
	
	if (strlen(url)==0)
		return 0
	endif
	
	fileType=parseFilePath(4, url,":", 0, 0)
	isZip=stringMatch(fileType, "zip")
	
	if(cmpstr(fileType, "ipf") && cmpstr(fileType, "zip")) 
		printf "New version of %s released\r", projectName
		printf "Updater couldn't find a useable file at %s\r", url
		return 0
	endif
	
	if (chatty)
		sprintf cmd, "%s: Found version %g. Download now?", projectName, releaseVersion
		doalert 1, cmd
		if (v_flag==2) // print file location for manual download
			printf "New version of %s available at %s\r", projectName, url
			return 0
		endif
	endif
		
	if(testing)
		// for testing, don't hit server pointlessly
		// instead, we'll 'download' a zip archive from the desktop
		printf "testing: substituting local file for %s\r", url
		url="file://"+specialDirPath("Desktop", 0, 1, 0)+"/archive.zip"
		if(isWin)
			url=replaceString("\\", url, "/")
		endif
		isZip=1
	endif	
	
	if(isZip==0) // an ipf file - download file and overwrite destination procedure	
		
		cmd=""
		if (!chatty) // in chatty mode this was already noted
			sprintf cmd, "New update found for %s\r", projectName
		endif	
		// Aways ask before overwriting file		
		sprintf cmd, "%sOverwrite %s with %s version %g?", cmd, ipfName, projectName, releaseVersion	
		doalert 1, cmd
		if(v_flag==2)
			printf "%s: update cancelled\r", projectName
			return 0
		endif
			
		if(strlen(backupPathStr)) // save a backup of the old procedure file
			// figure out full path to backup file
			sprintf backupPathStr, "%s%s%g.%s", backupPathStr, parseFilePath(3, procPathStr, ":", 0, 0), localVersion, parseFilePath(4, procPathStr, ":", 0, 0) 
			flagI=0
			getFileFolderInfo /Q/Z backupPathStr
			if(v_flag==0) // file already exists in archive location
				flagI=2 // check before overwriting
			endif
			movefile /O/S="Archive current file"/I=(flagI)/Z  procPathStr as backupPathStr
			if(v_flag==0)
				printf "Saved copy of %s version %g to %s\r", projectName, localVersion, s_path
			endif
		endif
				
		// download new ipf file and overwrite exisiting one
		URLRequest /TIME=(kFileTimeout)/Z/O/FILE=procPathStr url=url
		if(V_flag)
			printf "Could not download %s", url
			return 0
		elseif(strlen(S_fileName)==0)
			printf "Could not write file to %s", procPathStr
			return 0
		endif		
		printf "Downloaded %s\r", url
		printf "Saved file to %s\r", S_fileName
	endif
		
	if(isZip)
		// create a temporary folder in the download location
		newpath /C/O/Q tempPath, downloadPathStr; killpath /z tempPath
				
		// download zip file to temporary location
		URLRequest /TIME=(kFileTimeout)/Z/O/FILE=downloadPathStr+parseFilePath(0, url,"/", 1, 0) url=url
		if(V_flag)
			printf "Could not download %s", url
			return 0
		elseif(strlen(S_fileName)==0)
			printf "Could not write file to %s", procPathStr
			return 0
		endif
		
		archivePathStr=S_fileName // path to zip file
		archiveName=ParseFilePath(0, S_fileName, ":", 1, 0)
		printf "Downloaded %s version %g from %s\r", projectName, releaseVersion, url
		printf "Saved temporary file %s\r", archiveName
		
		// create a temporary directory for the uncompressed files
		string unzipPathStr=CreateUniqueDir(downloadPathStr, "Install")

		// inflate archive
		variable success = unzipArchive(archivePathStr, unzipPathStr)
		if (success==0)
			printf "unzipArchive failed to inflate %s\r", S_fileName
			return 0
		endif		
		printf "Inflated %s to %s\r", archiveName, unzipPathStr
		
		// get list of files and folders
		newpath /O/Q unzipPath, unzipPathStr
		fileList=IndexedFile(unzipPath, -1, "????")
		folderList=IndexedDir(unzipPath, -1, 0)
		killpath /Z unzipPath
			
		// ignore pesky __MACOSX folder, and other folders listed in ignoreFoldersList
		folderList=RemoveFromListWC(folderList, ksIgnoreFoldersList)	
		fileList=RemoveFromListWC(fileList, ksIgnoreFilesList)
		
		// use FindFile to look for the file that replaces the target procedure file
		string unzippedIPFpathStr=FindFile(unzipPathStr, ipfName)
		if(strlen(unzippedIPFpathStr)==0) // abort installation if we don't have the right file!
			printf "Could not find %s in %s\r", ipfName, archiveName
			deleteFolder /M="OK to delete temporary installation files?"/Z removeEnding(downloadPathStr,":")
			if(v_flag==0)
				printf "Deleted temporary folder %s\r", downloadPathStr
			endif
			return 0
		endif	
		
		// check number of files and folders at root of inflated archive
		variable numFiles=itemsinlist(fileList), numFolders=itemsinlist(folderList)
		
		if (numFiles==0 && numFolders==1)
			// inflated files are inside a subfolder of unzipPathStr
			unzipPathStr+=stringfromlist(0, folderList)+":"
			newpath /O/Q unzipPath, unzipPathStr
			fileList=IndexedFile(unzipPath, -1, "????")
			folderList=IndexedDir(unzipPath, -1, 0)
			killpath /Z unzipPath
			folderList=RemoveFromListWC(folderList, ksIgnoreFoldersList)	
			fileList=RemoveFromListWC(fileList, ksIgnoreFilesList)
			numFiles=itemsinlist(fileList)
			numFolders=itemsinlist(folderList)
		endif
					
		// figure out local path from unzipPathStr to replacement procedure file
		unzippedIPFpathStr=unzippedIPFpathStr[strlen(unzipPathStr),inf]
		variable subfolders=itemsinlist(unzippedIPFpathStr, ":")-1
			
		if(subfolders>0) // check that directory structure at destination matches that from archive
			if(stringmatch(procPathStr, "*"+unzippedIPFpathStr)==0)
				// target directory doesn't match folder in archive
				sprintf cmd, "Could not match directory structure in %s to %s\r", archiveName, packagePathStr
				print cmd
				if (chatty)
					doAlert 0, cmd
				endif
				deleteFolder /M="OK to delete temporary installation files?"/Z removeEnding(downloadPathStr,":")
				if(v_flag==0)
					printf "Deleted temporary folder %s\r", downloadPathStr
				endif
				return 0
			endif
		endif		
	
		// move up directory structure at destination
		destPathStr=ParseFilePath(1, procPathStr, ":", 1, 0) // path to folder containing file
		for(i=0;i<subfolders;i+=1)
			destPathStr=ParseFilePath(1, destPathStr, ":", 1, 0)
		endfor
		
		// test to find out which files will be overwritten
		fileList=mergeFolder(unzipPathStr, destPathStr, test=1)
		numFiles=ItemsInList(fileList)
		if (numFiles)
			if (!chatty) // in chatty mode this was already noted
				sprintf cmd, "New update found for %s\r", projectName
			endif
			cmd+="The following files will be overwritten:\r"
			for (i=0;i<numFiles;i+=1)
				fileName=stringFromList(i, fileList)
				fileName=fileName[strlen(SpecialDirPath("Igor Pro User Files",0,0,0)),inf]
				cmd+=fileName+"\r"
			endfor
			cmd+="Continue?"
			doAlert 1, cmd
			if(v_flag==2)
				printf "Update for %s cancelled\r", shortname
				deleteFolder /M="OK to delete temporary installation files?"/Z removeEnding(downloadPathStr,":")
				if(v_flag==0)
					printf "Deleted temporary folder %s\r", downloadPathStr
				endif
				return 0
			endif
		endif
		
		// create backup folder
		if(numFiles && strlen(backupPathStr))				
			string folderName=parseFilePath(3, procPathStr, ":", 0, 0) + num2str(localVersion)
			folderName=cleanupname(folderName,0) // making it Igor-legal should hopefully ensure the name is good for the OS
			sprintf backupPathStr, "%s%s:", backupPathStr, folderName
			NewPath /Q/C/O/Z tempPath, backupPathStr; killpath /Z tempPath
			printf "Created backup folder %s\r", backupPathStr
		endif
			
		// move files into user procedures folder
		fileList=mergeFolder(unzipPathStr, destPathStr, backupPathStr=backupPathStr, killSourceFolder=0)
		for (i=0;i<itemsinlist(fileList);i+=1)
			fileName=stringfromlist(i, fileList) // this is full path to file
			printf "Moved %s to %s\r", parsefilePath(0, fileName, ":", 1, 0), ParseFilePath(1, fileName, ":", 1, 0)
		endfor
		
		deleteFolder /M="OK to delete temporary installation files?"/Z removeEnding(downloadPathStr,":")
		if(v_flag==0)
			printf "Deleted temporary folder %s\r", downloadPathStr
		endif

	endif // end of download and inflate zip archive
	
	printf "Updated %s to version %g\r", projectName, releaseVersion
		
	string IgorProcPathStr=SpecialDirPath("Igor Pro User Files",0,0,0)+"Igor Procedures:"

	if(autoloadOption==1 && strlen(FindAlias(IgorProcPathStr, procPathStr))==0)
		sprintf cmd "Create %s in Igor Procedure to autoload %s?", selectstring(isWin, "alias", "shortcut"), projectName
		doAlert 1, cmd
		if (v_flag==1)
			createAliasShortcut procPathStr as IgorProcPathStr+ParseFilePath(3, procPathStr, ":", 0, 0)+selectstring(isWin, "Alias", "Shortcut") 
		endif			 
	endif
	
	return 1
end

static function /S CreateUniqueDir(pathStr, baseName)
	string pathStr, baseName
	
	pathStr=ParseFilePath(2, pathStr, ":", 0, 0)
	GetFileFolderInfo /Q/Z pathStr
	if(v_isFolder==0)
		return ""
	endif
	
	variable i=0
	string strOut=""
	do
		sprintf strOut,"%s%s%d:", pathStr, baseName, i
		GetFileFolderInfo /Q/Z strOut
		i+=1
	while(V_Flag==0)			
	newpath /C/O/Q tempPath, strOut; killPath /Z tempPath
	if (v_flag==0)
		return strOut
	endif
	return ""
end


// returns listStr, purged of any items that match an item in ZapListStr. 
// Wildcards okay! Case insensitive.
static function /S RemoveFromListWC(listStr, zapListStr)
	string listStr, zapListStr
	
	string removeStr=""
	variable i
	for(i=0;i<itemsinlist(zapListStr);i+=1)
		removeStr+=listmatch(listStr, stringfromlist(i, zapListStr))
	endfor
	return RemoveFromList(removeStr, listStr, ";", 0)
end

// Parse html in s_web to find releases. 
// currently we extract the info we need from columns 0, 1, 4 and 5
static function /WAVE ParseReleases(WebPageText)
	string WebPageText
	
	make /T/Free/N=(0,5) w_releases
		
	variable selStart=-1, selEnd=-1, blockStart, blockEnd, selFound
	string projectName="", requiredVersion="", releaseMajor="", releaseMinor="", url="", platform=""
	variable i=0
	
	// locate project name
	selStart=strsearch(WebPageText, "<h1 class=\"page-title\" title=\"Releases for ", blockStart, 2)
	selStart+=43
	selEnd=strsearch(WebPageText, "\"", selStart, 2)
	if(selStart<0 || selEnd>selStart+150)
		return w_releases
	endif
	projectName=WebPageText[selStart, selEnd-1]

	do
		// find start and end of project release fields
		selStart= strsearch(WebPageText, "<div class=\"project-release-info\">", selStart, 2)
		if(selStart==-1)
			break
		endif
		blockStart=selStart
		selEnd=strsearch(WebPageText, "<div class=\"project-release-footer\">", selStart, 2)
		if(selEnd==-1)
			break
		endif
		blockEnd=selEnd
		
		// locate version info
		selStart=strsearch(WebPageText, "\"field-paragraph-version\">", blockStart, 2)
		selStart+=26
		selEnd=strsearch(WebPageText, "<", selStart, 2)
		if(selStart<blockStart || selEnd>blockEnd)
			break
		endif
		requiredVersion=WebPageText[selStart, selEnd-1]	
		
		selStart=strsearch(WebPageText, "\"field-paragraph-version-major\">", blockStart, 2)
		selStart+=32
		selEnd=strsearch(WebPageText, "<", selStart, 2)
		if(selStart<blockStart || selEnd>blockEnd)
			break
		endif
		releaseMajor=WebPageText[selStart, selEnd-1]
		
		selStart=strsearch(WebPageText, "\"field-paragraph-version-patch\">", blockStart, 2)
		selStart+=32
		selEnd=strsearch(WebPageText, "<", selStart, 2)
		if(selStart<blockStart || selEnd>blockEnd)
			break
		endif
		releaseMinor=WebPageText[selStart, selEnd-1]
		
		// find download link
		selStart=strsearch(WebPageText, "\"field-paragraph-file\"", blockStart, 2)
		selStart=strsearch(WebPageText, "<a href=\"", selStart, 2)
		selStart+=9
		selEnd=strsearch(WebPageText, "\"", selStart, 2)
		if(selEnd<0 || selEnd>blockEnd)
			break
		endif
		url=WebPageText[selStart, selEnd-1]
				
		platform=""
		selFound=strsearch(WebPageText, "<span class=\"entity-reference\">Windows</span>", blockStart, 2)
		if(selFound>blockStart && selFound<blockEnd)
			platform+="Windows;"
		endif
		selFound=strsearch(WebPageText, "<span class=\"entity-reference\">Mac-", blockStart, 2)
		if(selFound>blockStart && selFound<blockEnd)
			platform+="Macintosh;"
		endif		
		if(strlen(projectName)==0 || strlen(releaseMajor)==0 || strlen(releaseMinor)==0 ||strlen(url)==0)
			break
		endif
		w_releases[dimsize(w_releases,0)][]={{projectName},{requiredVersion},{releaseMajor},{releaseMinor},{url},{platform}}
		selStart=blockEnd
		i+=1
	while (i<1000)
	return w_releases	
end

// extract a static string constant from a procedure file by reading from disk
static function /T getUpdaterStringConst(constantNameStr, pathToFile)
	string constantNameStr, pathToFile
	
	string s_exp="", s_out=""
	sprintf s_exp, "(?i)^[\s]*static[\s]*strconstant[\s]*%s[\s]*=", constantNameStr	
	Grep /Q/E=s_exp/LIST/Z pathToFile
	variable start, stop
	start=strsearch(s_value, "\"", 0)
	stop=strsearch(s_value, "\"", start+1)
	if (start<0 || stop<0)
		return ""
	endif
	return s_value[start+1,stop-1]
end

// strip trailing stuff in the output from winlist
static function /S extractFilename(s)
	string s
	
	SplitString/E=".*\.ipf" s // remove module names
	return removeending(s_value, ".ipf")
end
 
// utility function, inflates a zip archive
static function unzipArchive(archivePathStr, unzipPathStr)
	string archivePathStr, unzipPathStr
		
	string validExtensions="zip;" // set to "" to skip check
	variable verbose=1 // choose whether to print output from executescripttext
	string msg, unixCmd, cmd
		
	GetFileFolderInfo /Q/Z archivePathStr

	if(V_Flag || V_isFile==0)
		printf "Could not find file %s\r", archivePathStr
		return 0
	endif

	if(itemsInList(validExtensions) && findlistItem(ParseFilePath(4, archivePathStr, ":", 0, 0), validExtensions, ";", 0, 0)==-1)
		printf "%s doesn't appear to be a zip archive\r", ParseFilePath(0, archivePathStr, ":", 1, 0)
		return 0
	endif
	
	if(strlen(unzipPathStr)==0)
		unzipPathStr=SpecialDirPath("Desktop",0,0,0)+ParseFilePath(3, archivePathStr, ":", 0, 0)
		sprintf msg, "Unzip to %s:%s?", ParseFilePath(0, unzipPathStr, ":", 1, 1), ParseFilePath(0, unzipPathStr, ":", 1, 0)
		doALert 1, msg
		if (v_flag==2)
			return 0
		endif
	else
		GetFileFolderInfo /Q/Z unzipPathStr
		if(V_Flag || V_isFolder==0)
			sprintf msg, "Could not find unzipPathStr folder\rCreate %s?", unzipPathStr
			doalert 1, msg
			if (v_flag==2)
				return 0
			endif
		endif	
	endif
	
	// make sure unzipPathStr folder exists - necessary for mac
	newpath /C/O/Q acw_tmpPath, unzipPathStr
	killpath /Z acw_tmpPath

	if(stringmatch(StringByKey("OS", igorinfo(3))[0,2],"Win")) // Windows
		// The following works with .Net 4.5, which is available in Windows 8 and up.
		// current versions of Windows with Powershell 5 can use the more succinct PS command 
		// 'Expand-Archive -LiteralPath C:\archive.zip -DestinationPath C:\Dest'
		string strVersion=StringByKey("OSVERSION", igorinfo(3))
		variable WinVersion=str2num(strVersion) // turns "10.1.2.3" into 10.1
		if (WinVersion<8)
			print "unzipArchive requires Windows 8 or later"
			return 0
		endif
		
		archivePathStr=parseFilePath(5, archivePathStr, "\\", 0, 0)
		unzipPathStr=parseFilePath(5, unzipPathStr, "\\", 0, 0)
		cmd="powershell.exe -nologo -noprofile -command \"& { Add-Type -A 'System.IO.Compression.FileSystem';"
		sprintf cmd "%s [IO.Compression.ZipFile]::ExtractToDirectory('%s', '%s'); }\"", cmd, archivePathStr, unzipPathStr
	else // Mac
		sprintf unixCmd, "unzip %s -d %s", ParseFilePath(5, archivePathStr, "/", 0,0), ParseFilePath(5, unzipPathStr, "/", 0,0)
		sprintf cmd, "do shell script \"%s\"", unixCmd
	endif
	
	executescripttext /B/UNQ/Z cmd
	if(verbose)
		print S_value // output from executescripttext
	endif
	
	return (v_flag==0)
end

// returns full path to alias if found, otherwise ""
static function /T FindAlias(folderPathStr, targetPathStr)
	String folderPathStr, targetPathStr
	
	variable maxLevels=5 // quit after recursion through this many sublevels
	variable folderLimit=20 // quit after looking in this many folders.
	Variable folderIndex, fileIndex, subfolderIndex, folderCount=1, sublevels=0
	string folderList

	if (strlen(folderPathStr)==0 || strlen(targetPathStr)==0)
		return ""
	endif
		
	make /free/T/n=0 w_folders, w_subfolders
	w_folders={folderPathStr}

	do
		for(folderIndex=0;folderIndex<numpnts(w_folders);folderIndex+=1)
			// check files at current level
			newPath /O/Q/Z tempPath, w_folders[folderIndex]
			fileindex=0
			do
				getFileFolderInfo /P=tempPath/Q/Z IndexedFile(tempPath, fileindex, "????")
				if(V_isFile==0)
					break
				endif
				if(V_isAliasShortcut && stringmatch(S_aliasPath, targetPathStr))
					killpath /Z tempPath
					return s_path // set by getFileFolderInfo
				endif
				fileindex+=1
			while(1)
			// make a list of subfolders in current folder
			folderList=IndexedDir(tempPath, -1, 0)
			for (subfolderIndex=0;subfolderIndex<itemsinList(folderList);subfolderIndex+=1)
				w_subfolders[numpnts(w_subfolders)]={w_folders[folderIndex]+stringfromlist(subfolderIndex,folderList)+":"}				
				folderCount+=1
			endfor			
		endfor
		duplicate /T/O/free w_subfolders, w_folders
		redimension /N=0 w_subfolders
		sublevels+=1
		
		if(numpnts(w_folders)==0)
			break
		endif
		
		if(sublevels>maxLevels || (folderCount-folderIndex)>folderLimit)
			break
		endif

	while(1)
	killpath /Z tempPath
	return ""
end

// Search recursively for fileName within folder described by folderPathStr.
// Don't expect this to be quick.
// Returns path to file. Wildcards okay.
static function /T FindFile(folderPathStr, fileName)
	String folderPathStr, fileName
	
	variable maxLevels=10 // quit after recursion through this many sublevels
	variable folderLimit=5000 // quit after looking in this many folders.
	Variable folderIndex, fileIndex, subfolderIndex, folderCount=1, sublevels=0
	string folderList, indexFileName

	if (strlen(folderPathStr)==0)
		// it would be nice to default to the current save location, but I couldn't find a way to do that
		folderPathStr=SpecialDirPath("Documents", 0, 0, 0)
	endif
		
	make /free/T/n=0 w_folders, w_subfolders
	w_folders={folderPathStr}

	do
		for(folderIndex=0;folderIndex<numpnts(w_folders);folderIndex+=1)
			// check files at current level
			newPath /O/Q/Z tempPathFF, w_folders[folderIndex]

			fileindex=0
			do	
				indexFileName=IndexedFile(tempPathFF, fileindex, "????")
				if(stringmatch(indexFileName, fileName))
					killpath /Z tempPathFF
					return w_folders[folderIndex]+indexFileName
				endif
				fileindex+=1
			while(strlen(indexFileName))
			// make a list of subfolders in current folder
			folderList=IndexedDir(tempPathFF, -1, 0)
			for (subfolderIndex=0;subfolderIndex<itemsinList(folderList);subfolderIndex+=1)
				w_subfolders[numpnts(w_subfolders)]={w_folders[folderIndex]+stringfromlist(subfolderIndex,folderList)+":"}				
				folderCount+=1
			endfor			
		endfor
		duplicate /T/O/free w_subfolders, w_folders
		redimension /N=0 w_subfolders
		sublevels+=1
		
		if(numpnts(w_folders)==0)
			break
		endif
		
		if(sublevels>maxLevels || (folderCount-folderIndex)>folderLimit)
			break
		endif

	while(1)
	killpath /Z tempPathFF
	return ""
end

// search recursively in folderPathStr for files matching fileName.
// returns list of complete paths to files
static function /T RecursiveFileList(folderPathStr, fileName)
	String folderPathStr, fileName
	
	variable maxLevels=5 // quit after recursion through this many sublevels
	variable folderLimit=500 // quit after looking in this many folders
	Variable folderIndex, fileIndex, subfolderIndex, folderCount=1, sublevels=0
	string folderList, indexFileName
	string fileList=""

	if (strlen(folderPathStr)==0)
		return ""
	endif
		
	make /free/T/n=0 w_folders, w_subfolders
	w_folders={folderPathStr}

	do
		for(folderIndex=0;folderIndex<numpnts(w_folders);folderIndex+=1)
			// check files at current level
			newPath /O/Q/Z tempPathFF, w_folders[folderIndex]

			fileindex=0
			do	
				indexFileName=IndexedFile(tempPathFF, fileindex, "????")
				if(stringmatch(indexFileName, fileName))
					fileList=AddListItem(w_folders[folderIndex]+indexFileName, fileList)
				endif
				fileindex+=1
			while(strlen(indexFileName))
			// make a list of subfolders in current folder
			folderList=IndexedDir(tempPathFF, -1, 0)
			for (subfolderIndex=0;subfolderIndex<itemsinList(folderList);subfolderIndex+=1)
				w_subfolders[numpnts(w_subfolders)]={w_folders[folderIndex]+stringfromlist(subfolderIndex,folderList)+":"}				
				folderCount+=1
			endfor			
		endfor
		duplicate /T/O/free w_subfolders, w_folders
		redimension /N=0 w_subfolders
		sublevels+=1
		
		if(numpnts(w_folders)==0)
			break
		endif
		
		if(((maxlevels && sublevels>maxLevels) || (folderLimit && (folderCount-folderIndex)>folderLimit)))
			break
		endif

	while(1)
	killpath /Z tempPathFF
	return fileList
end

// Merges folders like copyFolder on Windows, deletes source files
// Files in destination folder are overwritten by files from source
// Setting test=1 doesn't move anything but generates a list of files 
// that would be overwritten
static function /S mergeFolder(source, destination, [killSourceFolder, backupPathStr, test])
	string source, destination
	variable killSourceFolder, test
	string backupPathStr // must be no more than one sublevel below an existing folder
	
	killSourceFolder = ParamIsDefault(killSourceFolder) ? 1 : killSourceFolder
	test = ParamIsDefault(test) ? 0 : test
	if(ParamIsDefault(backupPathStr))
		backupPathStr=""
	endif
	variable backup=(strlen(backupPathStr)>0)
	
	// clean up paths
	source=ParseFilePath(2, source, ":", 0, 0)
	destination=ParseFilePath(2, destination, ":", 0, 0)
	if (backup)
		backupPathStr=ParseFilePath(2, backupPathStr, ":", 0, 0)
	endif
	
	// check that souce and destination folders exist
	GetFileFolderInfo /Q/Z source
	variable sourceOK=V_isFolder
	GetFileFolderInfo /Q/Z destination
	if(sourceOK==0 || V_isFolder==0)
		return ""
	endif
	
	Variable folderIndex, fileIndex, subfolderIndex, folderCount=1, sublevels=0
	string folderList, fileList, fileName
	string movedFileList="", destFolderStr=""
	string subPathStr=""
	
	make /free/T/n=0 w_folders, w_subfolders
	w_folders={source}
	do
		// step through folders at current sublevel
		for(folderIndex=0;folderIndex<numpnts(w_folders);folderIndex+=1)
			// figure out destination folder to match current source folder
			subPathStr=(w_folders[folderIndex])[strlen(source),strlen(w_folders[folderIndex])-1]
			destFolderStr=destination+subPathStr
			
			// make sure that folder exists at destination
			if(test==0)
				newPath /C/O/Q/Z tempPathFF, destination+subPathStr
				if (backup) 
					newPath /C/O/Q/Z tempPathFF, backupPathStr+subPathStr
				endif
			endif
					
			// get list of source files in indexth folder at current sublevel
			newPath /O/Q/Z tempPathFF, w_folders[folderIndex]
			fileList=IndexedFile(tempPathFF, -1, "????")
			// remove files from list if they match an entry in ignoreFileList
			fileList=RemoveFromListWC(fileList, ksIgnoreFilesList)
			// move files 
			for(fileindex=0;fileIndex<itemsinList(fileList);fileIndex+=1)
				FileName=stringFromList(fileindex, fileList)			
				if(test)
					GetFileFolderInfo /Q/Z destFolderStr+FileName
					if(v_flag==0 && v_isFile) // file is to be overwritten
						movedFileList=AddListItem(destFolderStr+FileName, movedFileList)
					endif
				else
					if (backup) // back up any files that are to be overwritten
						GetFileFolderInfo /Q/Z destFolderStr+FileName
						if(v_flag==0 && v_isFile) // file is to be overwritten
							//newPath /C/O/Q/Z tempPath, backupPathStr+subPathStr; killpath /Z tempPath
							MoveFile /Z/O destFolderStr+FileName as backupPathStr+subPathStr+FileName
						endif
					endif
					MoveFile /Z/O w_folders[folderIndex]+FileName as destFolderStr+FileName
					movedFileList=AddListItem(destFolderStr+FileName, movedFileList)
				endif
			endfor
			
			// make a list of subfolders in current folder
			folderList=IndexedDir(tempPathFF, -1, 0)
			
			// remove folders that we don't want to copy
			folderList=RemoveFromListWC(folderList, ksIgnoreFoldersList)
			
			// add the list of folders to the subfolders wave
			for (subfolderIndex=0;subfolderIndex<itemsinList(folderList);subfolderIndex+=1)
				w_subfolders[numpnts(w_subfolders)]={w_folders[folderIndex]+stringfromlist(subfolderIndex,folderList)+":"}	
				folderCount+=1
			endfor			
		endfor
		// prepare for next sublevel iteration
		duplicate /T/O/free w_subfolders, w_folders
		redimension /N=0 w_subfolders
		sublevels+=1
	
		if(numpnts(w_folders)==0)
			break
		endif

	while(1)
	killpath /Z tempPathFF
	
	if((test==0) && killSourceFolder)
		deleteFolder /Z removeEnding(source, ":")
	endif
	
	return sortlist(movedFileList)
end

// ------------------------- Project Installer ---------------------

// install a user-contributed package hosted on Wavemetrics.com
// packageList is a list of project short names.
// The project URL can be used in place of the short name.
// Looks for install.itx within package folder and executes if found.
// install.itx should create shortcuts for help files and XOPs as needed.
// Use installWindows.itx or installMacintosh.itx for OS specific installations.

// install("foo") installs IgorExchange project with shortname foo in a 
// location within the Igor User Files folder chosen by the user. 

// install("foo;bar;", location=SpecialDirPath("Igor Pro User 
// Files",0,0,0)+"User Procedures:") installs packages foo and bar in the
// User Procedures folder.

// if location is supplied for zip archive(s), subfolders may or may not 
// not be created, depending on the structure of the archive. 

// uses parseReleases, unzipArchive, findFile, findAlias, extractFilename 
// functions from Updater code.

function install(packageList, [location, gui])
	string packageList, location
	variable gui
	
	if(paramIsDefault(gui)) // gui is not intended to be set when running from commandline
		gui=0 // don't use doAlert unless necessary
	endif
	
	// where to save downloaded files - a temporary location
	// Safe to edit this, but stuff in this location will be overwritten or deleted!
	// This function attempts to delete the folder after installation.
	string downloadPathStr=SpecialDirPath("Desktop",0,0,0)+"ACW_PackageInstallerTemp:"
	variable testing=0 //, autoloadOption=0 
	// autoloadOption=1 will prompt user to create an alias in Igor Procedures folder.
	
	variable removeSuffix=1 // if a single ipf is downloaded that has a name that looks like the 
 								// project shortname with a suffix to make the file unique, remove that suffix.
	
	string url, projectName, cmd
	string IncludeProcStr="", packagePathStr="", packageName=""
	variable selStart, selEnd, releaseIgorVersion, releaseVersion, refNum, err
	variable i, ListIndex=0
	
	variable currentIgorVersion=str2num(StringByKey("IGORFILEVERSION", igorinfo(3)))
	string system=selectString(stringmatch(StringByKey("OS", igorinfo(3))[0,2],"Mac"), "Windows", "Macintosh")
	variable isWin=stringmatch(system[0,2],"Win")
	variable nameIsURL=0
	
	for(ListIndex=0;ListIndex<itemsinList(packageList);ListIndex+=1)
		packageName=stringfromlist(ListIndex, packageList)
		
		// packageName can be package shortname or package URL
		if (stringmatch(packageName, "https://www.wavemetrics.com/*"))
			nameIsURL=1
			url=packageName
		else
			sprintf url, "https://www.wavemetrics.com/project/%s", packageName
		endif
		
		// check the project page on IgorExchange	
		URLRequest /TIME=(kPageTimeout)/Z url=url
		if(V_flag)
			sprintf cmd, "Installer could not load %s\r", url
			InstallFailMsg(cmd, gui)			
			continue
		endif
			
		if(nameIsURL)
			selStart=strsearch(S_serverResponse, "<link rel=\"canonical\" href=\"", 0, 0)
			selStart+=28
			selEnd=strsearch(S_serverResponse, "\"", selStart, 0)
			if (selEnd==-1 || selStart==-1)
				sprintf cmd, "Installer could not find releases at %s\r", url
				InstallFailMsg(cmd, gui)
				continue
			endif
			url=S_serverResponse[selStart,selEnd-1]
			packageName=parseFilePath(0, url, "/", 1, 0)
		endif
				
		printf "Found %s at %s\r", packageName, url
		
		// find the link to the 'all releases' page for this project
		selEnd=strsearch(S_serverResponse, ">View all releases</a>", 0, 2)
		selStart=strsearch(S_serverResponse, "href=", selEnd, 3)
		if (selEnd==-1 || selStart==-1)
			sprintf cmd, "Installer could not find releases at %s\r", url
			InstallFailMsg(cmd, gui)	
			continue
		endif
		url="https://www.wavemetrics.com"+S_serverResponse[selStart+6,selEnd-2]
		
		// try to download 'all releases' page		
		URLRequest /TIME=(kPageTimeout)/Z url=url
		if(V_flag)
			sprintf cmd, "Could not load %s\r", url
			InstallFailMsg(cmd, gui)
			continue
		endif
		printf "Looking for releases at %s\r", url
		
		// parse page to populate w_releases
		wave /T w_releases=ParseReleases(S_serverResponse) // free text wave
		if (dimsize(w_releases,0)==0)
			sprintf cmd, "Could not find packages at %s.\r", url
			InstallFailMsg(cmd, gui)
			continue
		endif	
		
		// find most recent OS- and version-compatible release
		url=""
		for(i=0;i<dimsize(w_releases, 0);i+=1)
			projectName = w_releases[i][0]
			sscanf (w_releases[i][1])[5,inf], "%f.x-%f", releaseIgorVersion, releaseVersion
	
			if(currentIgorVersion<releaseIgorVersion)
				sprintf cmd, "%s %g requires Igor Pro version >= %g\r", projectName, releaseVersion, releaseIgorVersion
				InstallFailMsg(cmd, gui)
			elseif(FindListItem(system, w_releases[i][5])==-1)
				sprintf cmd, "%s %g not available for %s\r", projectName, releaseVersion, system
				InstallFailMsg(cmd, gui)
			else
				url=w_releases[i][4]
				break
			endif
		endfor
		if (strlen(url)==0)
			continue
		endif
						
		if(testing)
			// for testing, don't hit server pointlessly
			printf "testing: substituting local file for %s\r", url
			url="file://C:/Users/tony/Desktop/archive.zip"
			url="file:///Users/tony/Desktop/archive.zip"	
		endif
		
		// create the temporary folder
		newpath /C/O/Q DLPathTemp, downloadPathStr; killpath /z DLPathTemp
		
		// download package
		URLRequest /TIME=(kFileTimeout)/Z/O/FILE=downloadPathStr+parseFilePath(0, url,"/", 1, 0) url=url
		if(V_flag)
			printf "Could not load %s", url
			continue
		endif
		string fileName=S_fileName
		printf "Downloaded %s\r", url
		printf "Saved temporary file %s\r", fileName
		
		// determine destination for file(s)
		if(paramisdefault(location))	
			packagePathStr=ChooseInstallLocation(packageName)
			if (strlen(packagePathStr)==0)
				continue
			endif	
		else
			getfileFolderInfo /Q/Z location
			if(v_flag!=0 || V_isFolder==0)
				sprintf cmd, "Install path not found: %s\r", location
				InstallFailMsg(cmd, gui)
				continue
			endif
			packagePathStr=location
		endif
		
		string fileList="" // list of files moved to User Procs folder
				
		// consider only two possibilities: an uncompressed ipf file or a zip archive
	
		if(stringmatch(parseFilePath(4, fileName,":", 0, 0), "ipf"))
			string destFileName
			
			destFileName=packagePathStr+parseFilePath(0, fileName, ":", 1, 0)
			
			if(removeSuffix) // this will force the file name to match project shortname.
				destFileName=parseFilePath(3, fileName, ":", 0, 0)
				if(stringmatch(destFileName,packageName+"*"))
					destFileName=destFileName[0,strlen(packageName)-1] // preserve case of filename
					destFileName=packagePathStr+destFileName+".ipf"
				endif
			endif
			
			getFileFolderInfo /Q/Z destFileName
			if (v_flag==0) // already exists
				doalert 1, "overwrite "+parseFilePath(0, destFileName, ":", 1, 0)+"?"
				if (v_flag==2)
					continue
				endif
			endif
			movefile /O/Z 	fileName as destFileName
			if(V_flag!=0)
				continue
			endif
			fileList=destFileName
			printf "Saved %s in %s\r", parseFilePath(0, destFileName, ":", 1, 0), packagePathStr
		endif		
		
		if (stringmatch(parseFilePath(4, fileName,":", 0, 0), "zip"))
			//string archivePathStr=fileName // path to zip file
			string archiveName=ParseFilePath(0, fileName, ":", 1, 0)

			// create a temporary directory for the uncompressed files
			string unzipPathStr=CreateUniqueDir(downloadPathStr, "Install")

			// inflate archive
			variable success = unzipArchive(fileName, unzipPathStr)
			if (success==0)
				printf "unzipArchive failed to inflate %s\r", fileName
				return 0
			endif		
			printf "Inflated %s to %s\r", archiveName, unzipPathStr

			// move files/folders to destination	
			// test to find out which files will overwritten
			fileList=mergeFolder(unzipPathStr, packagePathStr, test=1)
			variable numFiles=ItemsInList(fileList)
			if (numFiles)
				cmd="The following files will be overwritten:\r"
				for (i=0;i<numFiles;i+=1)
					fileName=stringFromList(i, fileList)
					fileName=fileName[strlen(SpecialDirPath("Igor Pro User Files",0,0,0)),inf]
					cmd+=fileName+"\r"
				endfor
				cmd+="Continue?"
				doAlert 1, cmd
				if(v_flag==2)
					printf "Install cancelled for %s\r", packageName
					deleteFolder /M="OK to delete temporary installation files?"/Z removeEnding(downloadPathStr,":")
					if(v_flag==0)
						printf "Deleted temporary folder %s\r", downloadPathStr
					endif
					continue
				endif
			endif
			
			// move files into user procedures folder
			fileList=mergeFolder(unzipPathStr, packagePathStr, killSourceFolder=0)
			for (i=0;i<itemsinlist(fileList);i+=1)
				fileName=stringfromlist(i, fileList) // this is full path to file
				printf "Moved %s to %s\r", parsefilePath(0, fileName, ":", 1, 0), ParseFilePath(1, fileName, ":", 1, 0)
			endfor
		endif // end of download and inflate zip archive
		
		
		// figure out if we have a 'master' procedure file
		string ipfList=listMatch(fileList, "*.ipf")
		variable numIPFs=itemsInList(ipfList)	
		if(numIPFs==1) // only one procedure file
			IncludeProcStr=ParseFilePath(3, StringFromList(0, ipfList), ":", 0, 0)
		elseif(numIPFs>1 && (WhichListItem(packagePathStr+packageName+".ipf", ipfList)!=-1) )
			// package folder includes a procedure with project name
			IncludeProcStr=packageName
		endif
		// if there's an XOP in the package, don't try to open any procedure file
		if(strlen(listMatch(fileList, "*.xop")))
			IncludeProcStr=""
		endif
						
		// Look for and run itx-format installation script. Script will run
		// from within the package folder saved in location of user's 
		// choice within Igor Pro User Files folder. 
		string msg
		string pathToScriptStr=listMatch(fileList, "*:install.itx")
		if(strlen(pathToScriptStr)==0)
			pathToScriptStr=listMatch(fileList, "*:install"+system+".itx")
		endif
		if(strLen(pathToScriptStr)) // script could do naughty things, so ask before allowing it to run
			sprintf msg, "Run installation script?\r%s was downloaded from %s", ParseFilePath(0, pathToScriptStr, ":", 1, 0), url 
			doAlert 1, msg
			if (v_flag==1)
				sprintf cmd, "Loadwave /T \"%s\"", s_path
				execute cmd // will be printed to history
			endif
		endif
	
		if (strlen(IncludeProcStr)) // found a 'master' procedure file			
//			if(autoloadOption==1)
//				// find out whether an alias already exists
//				string IgorProcPathStr=SpecialDirPath("Igor Pro User Files",0,0,0)+"Igor Procedures:"
//				string targetPathStr=packagePathStr+IncludeProcStr+".ipf"
//				autoloadOption=(strlen(FindAlias(IgorProcPathStr, targetPathStr))==0)
//				// make sure target isn't in Igor Procedures folder
//				if(autoloadOption && stringmatch(targetPathStr, IgorProcPathStr+"*")==0)
//					string aliasName=ParseFilePath(3, targetPathStr, ":", 0, 0)+selectstring(isWin, " Alias", " Shortcut")
//					sprintf cmd "Create %s alias in Igor Procedures to autoload %s?", ParseFilePath(0, targetPathStr, ":", 1, 0), packageName
//					doAlert 1, cmd
//					if (v_flag==1)
//						createAliasShortcut targetPathStr as IgorProcPathStr+aliasName
//						printf "Created %s in %s\r", aliasName, IgorProcPathStr
//					endif 
//				endif
//			endif
			
			// find out if procedure is already loaded
			wave /T w_loadedProcs=ListToTextWave(winlist("*",";","WIN:128,INDEPENDENTMODULE:1"),";")
			if(numpnts(w_loadedProcs))
				w_loadedProcs=extractFilename(w_loadedProcs)
			endif
			FindValue /TEXT=IncludeProcStr/TXOP=4 /Z w_loadedProcs
			
			if(V_Value==-1)
				sprintf cmd "Do you want to include %s in the current experiment?", IncludeProcStr
				doAlert 1, cmd
				if(v_flag==1)
					// load the procedure
					sprintf cmd, "INSERTINCLUDE \"%s\"", IncludeProcStr
					Execute/P/Q/Z cmd
					Execute/P/Q/Z "COMPILEPROCEDURES "
					printf "Included %s in current experiment\r", IncludeProcStr
				endif	
			endif
						
		endif
	endfor

	deleteFolder /M="OK to delete temporary installation files?"/Z removeEnding(downloadPathStr,":")
	if(v_flag==0)
		printf "Deleted temporary folder %s\r", downloadPathStr
	endif
	
	return 1	
end

static function InstallFailMsg(cmd, gui)
	string cmd
	variable gui
	
	print cmd
	if(gui)
		doalert 0, cmd
	endif
	return 1
end

// returns a path to a subfolder of Igor Pro User Files selected by user
static function /S ChooseInstallLocation(packageName)
	string packageName
		
	variable success
	string cmd
	if(strlen(packageName))
		packageName=" "+packageName
	endif
	sprintf cmd, "Where would you like to install%s?", packageName
	do
		success=1
		NewPath /O/Q/Z TempInstallPath SpecialDirPath("Igor Pro User Files",0,0,0)+"User Procedures:"
		pathinfo /S TempInstallPath
		NewPath /M=cmd /O/Q/Z TempInstallPath
		if(v_flag)
			printf "Install cancelled\r"
			return ""
		endif
		pathinfo /S TempInstallPath
		if(stringmatch(S_path, SpecialDirPath("Igor Pro User Files",0,0,0)+"*")==0)
			doalert 0, "Please select a location within the Igor Pro User Files folder\r(usually in the User Procedures folder)"
			success=0
		endif
		if(stringmatch(S_path, SpecialDirPath("Igor Pro User Files",0,0,0)))
			doalert 0, "Please select a subfolder within the Igor Pro User Files folder\r(usually in the User Procedures folder)"
			success=0
		endif
	while(!success)
	killpath /Z TempInstallPath
	return S_path
end	



// -------------------- a GUI for installing user projects ------------

menu "Macros"
	"Install user project...", /Q, updater#makeInstallerPanel()
end

static function makeInstallerPanel()
	
	dowindow /k InstallerPanel
	
	NewDataFolder /O root:Packages
	NewDataFolder /O root:Packages:Installer
	DFREF dfr = root:Packages:Installer
	variable /G dfr:stubLen=0

	// create wave to hold names of all available projects
	wave /T/SDFR=dfr fullList
	if(waveexists(fullList)==0)
		make /O/T/N=(0,6) dfr:fullList /WAVE=fullList, dfr:matchList /WAVE=matchlist
		setDimLabel 1, 0, name, fullList, matchlist
		setDimLabel 1, 1, author, fullList, matchlist
		setDimLabel 1, 2, projectURL, fullList, matchlist
		setDimLabel 1, 3, type, fullList, matchlist
		setDimLabel 1, 4, published, fullList, matchlist
		setDimLabel 1, 5, views, fullList, matchlist
	endif
	
	make /o/n=1/T dfr:displayList /WAVE=displayList, dfr:helpList /WAVE=helpList
	displayList="retrieving list..."
	helpList="waiting for download"
	// for now, these are 1D waves... 
	// but duplicate will turn them into single column 2D waves
	// so be careful about indexing!
		
	// make a control panel GUI	
	NewPanel /K=1/W=(100,50,420,420)/N=InstallerPanel as "Install User Project"
	
	variable vLeft=10, vTop=10
	
	NewNotebook /F=1 /N=nb0 /HOST=InstallerPanel /W=(vLeft,vTop,295,45) /OPTS=3
	notebook InstallerPanel#nb0 fSize=12, showRuler=0
	SetWindow InstallerPanel#nb0, activeChildFrame=0
	
	Button buttonClear, win=InstallerPanel,pos={297,20},size={15,15},title=""
	Button buttonClear, picture=ProcGlobal#clearText,proc=Updater#InstallerButtonProc, disable=1
	vTop+=45
	
	ListBox listbox0, win=InstallerPanel, pos={vLeft,vTop},size={300, 200}, listwave=displayList
	ListBox listbox0, win=InstallerPanel, mode=1, proc=Updater#InstallerListBoxProc, selRow=-1
	ListBox listbox0, win=InstallerPanel, helpWave=helpList
	vTop+=210
	
	GroupBox group0, pos={vLeft,vTop}, size={300,70}
	
	popupMenu popupAuthor, win=InstallerPanel, pos={35,vTop+5}, value="all;", mode=1, title="   Author", proc=Updater#InstallerPopupProc
	popupMenu popupType, win=InstallerPanel, pos={35,vTop+25}, value="all;", mode=1, title="      Type", proc=Updater#InstallerPopupProc
	popupMenu popupOrder, win=InstallerPanel, pos={35,vTop+45}, value="project name;author;date;views", mode=3, title="Order by", proc=Updater#InstallerPopupProc
	
	vTop+=75
	Button buttonInstall, win=InstallerPanel,pos={240,vTop},size={60,20},proc=Updater#InstallerButtonProc,title="Install"
	Button buttonInstall, win=InstallerPanel,help={"Install selected user project"}, disable=2
	
	doupdate /W=InstallerPanel
	
	SetWindow InstallerPanel hook(hInstallerHook)=updater#InstallerPanelHook	

	GetProjects() // populate fullList
	if(dimsize(fullList,0)==0)
		displayList="Download failed."
	endif
	
	Duplicate /free/T /RMD=[][1,1] fullList, authors
	string authorList = "\"all;"+listOf(authors)+"\""
	popupMenu popupAuthor, win=InstallerPanel, value=#authorList, mode=1
	
	Duplicate /free/T /RMD=[][3,3] fullList, types
	string typeList = "\"all;"+listOf(types)+"\""
	popupMenu popupType, win=InstallerPanel, value=#typeList, mode=1
	
end

// get a list of user contributed projects from Wavemetrics.com
static function GetProjects()
	
	// get project name, author, and URL for each user contribution	
	DFREF dfr = root:Packages:Installer
	wave /T/SDFR=dfr fullList, displayList
	
	variable lastmod=NumberByKey("MODTIME", waveInfo(fullList,0))
	
	if(((datetime-lastmod) < 86400) && (dimsize(fullList, 0)>100)) // less than 1 day old
		// looks like we already have an up to date list of projects
		
		// make sure list is in default order
		make /free/n=(dimsize(fullList, 0)) sortwave
		sortwave=parseDate(fullList[p][%published])
		SortColumns /R keywaves={sortwave}, sortwaves={fullList}		
		
		UpdateListboxWave("")
		return 0
	endif
	
	// clear any incomplete or old list
	redimension /N=(0,6) fullList
	
	string baseURL, projectsURL, URL //, listPageText, projectPageText
	baseURL="https://www.wavemetrics.com"
	projectsURL="/projects?os_compatibility=All&project_type=All&field_supported_version_target_id=All&page="
	variable pageNum, projectNum, success
	
	variable pStart, pEnd, selStart, selEnd=0, err
	string strAuthor, strName, strShortName, strProjectURL
	
	// loop through listPages
	for(pageNum=0;pageNum<100;pageNum+=1)
		sprintf URL "%s%s%d", baseURL, projectsURL, pageNum
		
		URLRequest /TIME=(kPageTimeout)/Z url=url
		if(V_flag)
			printf "Installer could not load %s\r", url
					
			UpdateListboxWave("") // update based on anything we managed to download
			return 0		
		endif
			
		pStart=strsearch(S_serverResponse, "<section class=\"project-teaser-wrapper\">", 0, 0)
		if (pStart==-1)
			break // no more projects
		endif
		pEnd=0
		
		// loop through projects on listPage
		for (projectNum=0;projectNum<50; projectNum+=1)
		
			pStart=strsearch(S_serverResponse, "<section class=\"project-teaser-wrapper\">", pEnd, 0)	
			pEnd=strsearch(S_serverResponse, "<div class=\"project-teaser-footer\">", pStart, 0)
			if (pEnd==-1 || pStart==-1)
				break // no more projects on this listPage
			endif
			
			selStart=strsearch(S_serverResponse, "<span class=\"username-wrapper\">", pEnd, 3)
			if(selStart<pStart)
				continue
			endif
		
			selStart=strsearch(S_serverResponse, "\">", selStart, 0)
			selStart+=2
			selEnd=strsearch(S_serverResponse, "</span>", selStart, 0)		
			strAuthor=S_serverResponse[selStart,selEnd-1]
					
			selStart=strsearch(S_serverResponse, "<a href=\"", selEnd, 0)
			selStart+=9
			selEnd=strsearch(S_serverResponse, "\"><h2>", selStart, 0)
			strProjectURL=baseURL+S_serverResponse[selStart,selEnd-1]
			
			selStart=selEnd+6
			selEnd=strsearch(S_serverResponse, "</h2></a>", selStart, 0)
			strName=S_serverResponse[selStart,selEnd-1]
			
			if (strLen(strName)==0)
				continue
			endif


// Don't try to get project details at this point: too many pages to download			
			
//			projectPageText=FetchURL(strProjectURL); err = GetRTError(1)
//			if (err != 0)
//				continue
//			endif
//			selStart=strsearch(projectPageText, "<link rel=\"canonical\" href=\"", 0, 0)
//			selStart+=28
//			selEnd=strsearch(projectPageText, "\"", selStart, 0)
//			strProjectURL=projectPageText[selStart,selEnd-1]
//			strShortName=parseFilePath(0, strProjectURL, "/", 1, 0)
//			
//			success=(strlen(strShortName)&&strlen(strProjectURL))
//			if(!success)
//				continue
//			endif
			
			InsertPoints /M=0 DimSize(fullList, 0), 1, fullList
			fullList[inf][%author]=strAuthor
			fullList[inf][%name]=strName
			fullList[inf][%ProjectURL]=strProjectURL
			
			// search for other non-essential parameters
			
			// project types
			selEnd=pStart
			do
				selStart=strsearch(S_serverResponse, "/taxonomy/", selEnd, 0)
				selStart=strsearch(S_serverResponse, ">", selStart, 0)
				selEnd=strsearch(S_serverResponse, "<", selStart, 0)
				if(selStart<pStart || selEnd>pEnd || selEnd<1 )
					break
				endif
				fullList[inf][%type]+=S_serverResponse[selStart+1,selEnd-1]+";"
			while (1)
			
			// date	
			selStart=strsearch(S_serverResponse, "<span>", pEnd, 0)
			selEnd=strsearch(S_serverResponse, "</span>", pEnd, 0)
			if (selStart>0 && selEnd>0 && selEnd < (pEnd+80) )
				fullList[inf][%published]=S_serverResponse[selStart+6,selEnd-1]
			endif
			
			// views
			selEnd=strsearch(S_serverResponse, " views</span>", pEnd, 0)
			selStart=strsearch(S_serverResponse, "<span>", selEnd, 1)		
			if (selStart>pEnd && selEnd < (pEnd+150) )
				fullList[inf][%views]=S_serverResponse[selStart+6,selEnd-1]
			endif
		endfor	
		
		UpdateListboxWave("") // rebuild display list each time a web page is parsed for projects
		doupdate
	endfor
	
	return 1
end

// returns alphabetical list of textWave contents
static function /S listOf(textWave)
	wave /T textWave
	
	string outputList="", strItem
	variable i, j, items
	for (i=0;i<dimsize(textWave,0);i+=1)
		items=itemsInList(textWave[i][0])
		for(j=0;j<items;j+=1)
			strItem=stringFromList(j, textWave[i][0])
			if (WhichListItem(strItem,outputList)==-1)
				outputList=addListItem(strItem, outputList)
			endif
		endfor
	endfor
	
	return sortList(outputList)
end

static function ParseDate(str)
	string str
	
	string months="January;February;March;April;May;June;July;August;September;October;November;December"
	string strMon, strDay, strYear
	SplitString/E="([[:alpha:]]+) ([[:digit:]]+), ([[:digit:]]+)" str, strMon, strDay, strYear
	variable year, month, day
	month=WhichListItem(strMon, months)+1
	day=str2num(strDay)
	year=str2num(strYear)
	
	return dateToJulian(year, month, day)
end

static function InstallerPopupProc(s)
	STRUCT WMpopupAction &s
	
	if(s.eventCode!=2)
		return 0
	endif
		
	if(stringMatch(s.ctrlName, "popupOrder"))
		wave /T fullList=root:Packages:Installer:fullList
		variable sortColumn=s.popNum
		if(s.popNum<3) // text sort based on project name (s.popNum=1) or author.
			SortColumns /KNDX={s.popNum-1} sortwaves={fullList}
		else
			make /free/n=(dimsize(fullList, 0)) sortwave
			if(s.popNum==3) // date
				sortwave=parseDate(fullList[p][%published])
			elseif(s.popNum==4) // views
				sortwave=str2num(fullList[p][%views])
				sortwave = numtype(sortwave)==2 ? 0 : sortwave
			endif
			SortColumns /R keywaves={sortwave}, sortwaves={fullList}
		endif
	endif
	
	UpdateListboxWave(getStub())	
end

static function InstallerButtonProc(s)
	STRUCT WMButtonAction &s
	
	if(s.eventCode!=2)
		return 0
	endif
	
	strswitch(s.ctrlName)
		case "buttonInstall" :
			InstallSelectionIfValid()
			break
		case "buttonClear" :
			Notebook InstallerPanel#nb0 selection={startOfFile,endOfFile}, text="" // clear text	
			NVAR stubLen=root:Packages:Installer:stubLen
			stubLen=0
			Button buttonInstall, win=InstallerPanel, disable=1
			Button buttonClear, win=InstallerPanel, disable=3
			UpdateListboxWave("")
			break
	endswitch

	return 0
end

static function InstallSelectionIfValid()
	
	DFREF dfr = root:Packages:Installer
	wave /SDFR=dfr/T  matchList
	
	GetSelection notebook, InstallerPanel#nb0, 1 // get current position in notebook
	Notebook InstallerPanel#nb0 selection={startOfFile,endOfFile}
	GetSelection notebook, InstallerPanel#nb0, 2
	Notebook InstallerPanel#nb0 selection={(0,V_startPos),(0,V_endPos)}
	
	findValue /RMD=[][0,0]/TEXT=s_selection /TXOP=4/Z matchList
	if(v_value>-1) 
		print "install(\""+matchList[v_value][%projectURL]+"\")"
		install(matchList[v_value][%projectURL], gui=1)
	endif
end

static function InstallerListBoxProc(s)
	STRUCT WMListboxAction &s
	
	if(s.eventCode==4 || s.eventCode==5) // a list item is selected
		if(s.row>=numpnts(s.listwave))
			return 0
		endif
		
		// insert text into notebook subwindow	
		Notebook InstallerPanel#nb0 selection={startOfFile,endOfFile}, textRGB=(0,0,0), text=s.listwave[s.row][0]
		NVAR stubLen=root:Packages:Installer:stubLen
		stubLen=strlen(s.listwave[s.row][0])
		Button buttonInstall, win=InstallerPanel, disable=0
		//UpdateListboxWave(s.listwave[s.row])
		//commented out so that user can click through the list of matches.
		
	endif
	return 0
end

// update listbox wave based on string str
static function UpdateListboxWave(str)
	string str
	
	DFREF dfr = root:Packages:Installer
	wave /SDFR=dfr/T  fullList, matchList
	duplicate /free/T fullList, subList
	
	string regEx
	
	controlinfo /W=InstallerPanel popupAuthor
	if (cmpstr(S_Value, "all"))
		regEx="\\b"+s_value+"\\b"
		grep /GCOL=1/Z/E=regEx subList as subList
	endif
	controlinfo /W=InstallerPanel popupType
	if (cmpstr(S_Value, "all"))
		regEx="\\b"+s_value+"\\b"
		grep /GCOL=3/Z/E=regEx subList as subList
	endif	
	
	regEx="(?i)"+str		
	grep /GCOL=0/Z/E=regEx subList as matchList
	
	if (dimsize(matchList,0)) // if matchlist has zero rows duplicate can cause Igor to become unstable
		duplicate /T/O/R=[][0] matchList, dfr:displayList
		duplicate /T/O/R=[][1] matchList, dfr:helpList
	else
		make /o/n=0/T dfr:displayList, dfr:helpList
	endif
		
	ListBox listbox0, win=InstallerPanel, selRow=-1
	
	FindValue /TEXT=str/TXOP=4 dfr:displayList
	Button buttonInstall, win=InstallerPanel, disable=2*(v_value<0)
end

// intercept and deal with keyboard events in notebook subwindow
static function InstallerPanelHook(s)
	STRUCT WMWinHookStruct &s
	
	DFREF dfr = root:Packages:Installer

// the following is commented so that we leave data folder intact
// that way we don't have to download project list every time we initialize

//	if(s.eventcode==2)
//		killDataFolder /Z dfr
//		return 0
//	endif
	
	GetWindow $s.winName activeSW
	if (CmpStr(S_value,"InstallerPanel#nb0") != 0)
		return 0
	endif
	
	NVAR stubLen=dfr:stubLen

	if(s.eventcode==5) // mouseup 	
		GetSelection notebook, InstallerPanel#nb0, 1 // get current position in notebook
		V_endPos=min(stubLen,V_endPos)
		V_startPos=min(stubLen,V_startPos)
		Notebook InstallerPanel#nb0 selection={(0,V_startPos),(0,V_endPos)}
		return 1
	endif
	
	if(s.eventcode==10) // menu 
		strswitch(s.menuItem)
			case "Paste":
				GetSelection notebook, InstallerPanel#nb0, 1 // get current position in notebook		
				string strScrap=getscrapText()
				strScrap=replacestring("\r", strScrap, "")
				strScrap=replacestring("\n", strScrap, "")
				strScrap=replacestring("\t", strScrap, "")
				Notebook InstallerPanel#nb0 selection={(0,V_startPos),(0,V_endPos)}, text=strScrap
				stubLen+=strlen(strScrap)-abs(V_endPos-V_startPos)
				s.eventcode=11 
				// pretend this was a keyboard event to allow execution to continue
				break
			case "Cut":
				GetSelection notebook, InstallerPanel#nb0, 3 // get current position in notebook
				putscrapText s_selection
				Notebook InstallerPanel#nb0 selection={(0,V_startPos),(0,V_endPos)}, text=""
				stubLen-=strlen(s_selection)
				s.eventcode=11 
				break
			case "Clear":
				GetSelection notebook, InstallerPanel#nb0, 3 // get current position in notebook
				Notebook InstallerPanel#nb0 selection={(0,V_startPos),(0,V_endPos)}, text="" // clear text
				stubLen-=strlen(s_selection)
				s.eventcode=11 
				break
		endswitch
		Button buttonClear, win=InstallerPanel, disable=3*(stublen==0)
	endif
		
	if(s.eventcode!=11)		
		return 0
	endif
	
	// deal with some non-printing characters
	switch(s.keycode)
		case 9:	// tab: jump to end
		case 3:
		case 13: // enter or return: jump to end
			notebook InstallerPanel#nb0 selection={startOfFile,endofFile}, textRGB=(0,0,0)
			notebook InstallerPanel#nb0 selection={endOfFile,endofFile}
			GetSelection notebook, InstallerPanel#nb0, 1 // get current position in notebook
			stubLen=V_endPos
			break
		case 28: // left arrow
			return 0
		case 29: // right arrow
			GetSelection notebook, InstallerPanel#nb0, 1
			if(V_endPos>=stubLen)
				if(s.eventMod&2) // shift key
					Notebook InstallerPanel#nb0 selection={(0,V_startPos),(0,stubLen)}
				else
					Notebook InstallerPanel#nb0 selection={(0,stubLen),(0,stubLen)}
				endif
				return 1
			endif
			return 0
		case 8:
		case 127: // delete or forward delete
			GetSelection notebook, InstallerPanel#nb0, 1
			if(V_startPos==V_endPos)
				V_startPos += (s.keycode==8) ? -1 : 1
			endif
			V_startPos=min(stubLen,V_startPos); V_endPos=min(stubLen,V_endPos)
			V_startPos=max(0, V_startPos); V_endPos=max(0, V_endPos)
			Notebook InstallerPanel#nb0 selection={(0,V_startPos),(0,V_endPos)}, text=""
			stubLen-=abs(V_endPos-V_startPos)
			break			
	endswitch
		
	// find and save current position	
	GetSelection notebook, InstallerPanel#nb0, 1
	variable selEnd=V_endPos
		
	if(strlen(s.keyText)==1) // a one-byte printing character
		// insert character into current selection
		Notebook InstallerPanel#nb0 text=s.keyText, textRGB=(0,0,0)
		stubLen+=1-abs(V_endPos-V_startPos)
		// find out where we want to leave cursor
		GetSelection notebook, InstallerPanel#nb0, 1 
		selEnd=V_endPos
	endif	
	
	string strStub="", strInsert="", strEnding=""
		
	// select and format stub
	Notebook InstallerPanel#nb0 selection={startOfFile,(0,stubLen)}, textRGB=(0,0,0)	
	// get stub text
	GetSelection notebook, InstallerPanel#nb0, 3
	strStub=s_selection
	// get matches based on stub text
	UpdateListboxWave(strStub)
	
	// do auto-completion based on stubLen characters
	wave /T  matchList=dfr:matchList
	
	if(s.keycode==30 || s.keycode==31) // up or down arrow
		Notebook InstallerPanel#nb0 selection={(0,stubLen),endOfFile}
		GetSelection notebook, InstallerPanel#nb0, 3
		strEnding=s_selection
		strInsert=arrowKey(strStub, strEnding, 1-2*(s.keycode==30), matchList)
	else
		strInsert=completeStr(strStub, matchList)
	endif
	// insert completion text in grey	
	Notebook InstallerPanel#nb0 selection={(0,stubLen),endOfFile}, textRGB=(50000,50000,50000), text=strInsert	
	Notebook InstallerPanel#nb0 selection={(0,selEnd),(0,selEnd)}, findText={"",1}
	
	Button buttonClear, win=InstallerPanel, disable=3*(stublen==0)
	
	return 1 // tell Igor we've handled all keyboard events
end

static function /T getStub()
	
	DFREF dfr = root:Packages:Installer
	NVAR stubLen=dfr:stubLen
	
	// find and save current position	
	GetSelection notebook, InstallerPanel#nb0, 1
	variable selEnd=V_endPos
	
	// select stub
	Notebook InstallerPanel#nb0 selection={startOfFile,(0,stubLen)}
	
	// get stub text
	GetSelection notebook, InstallerPanel#nb0, 3
	string strStub=s_selection
	
	// reset position
	Notebook InstallerPanel#nb0 selection={(0,selEnd),(0,selEnd)}, findText={"",1}
	
	return strStub
end

// returns completion text for first match of string s in text wave w
static function /T completeStr(s, w)
	string s
	wave /T w
	
	if (strlen(s)==0)
		return ""
	endif
	
	string stub=s[0,strlen(s)-1]
	variable stubLen=strlen(stub)
	if (stubLen==0)
		return ""
	endif
	
	make /free/T/N=1 w_out
	grep /Z/E="(?i)^"+stub w as w_out
	if(dimsize(w_out,0)==0)
		return ""
	endif
	return (w_out[0][0])[stubLen,inf]	
end

// find next or previous matching entry in wList and return completion text
static function /T arrowKey(stub, ending, increment, wList)
	string stub, ending
	variable increment
	wave /T wList
	
	if (strlen(stub)==0)
		return ""
	endif
	
	variable stubLen=strlen(stub)
	if (stubLen==0)
		return ""
	endif
	
	make /free/T/N=1 w_out
	grep /Z/E="(?i)^"+stub/DCOL={0} wList as w_out
	if(dimsize(w_out,0)==0)
		return ""
	endif
	
	findValue /RMD=[][0,0]/TEXT=stub+ending /TXOP=4/Z w_out
	if(v_value>-1)
		v_value+=increment
		v_value = V_value<0 ? DimSize(w_out,0)-1 : v_value
		v_value = V_value>=DimSize(w_out,0) ? 0 : v_value
	else
		return (w_out[0][0])[stubLen,inf]
	endif
	return (w_out[v_value][0])[stubLen,inf]	
end

// ----------------------------------------------------------------

// PNG: width= 90, height= 30
Picture clearText
	ASCII85Begin
	M,6r;%14!\!!!!.8Ou6I!!!"&!!!!?#R18/!3BT8GQ7^D&TgHDFAm*iFE_/6AH5;7DfQssEc39jTBQ
	=U"$&q@5u_NKm@(_/W^%DU?TFO*%Pm('G1+?)0-OWfgsSqYDhC]>ST`Z)0"D)K8@Ncp@>C,GnA#([A
	Jb0q`hu`4_P;#bpi`?T]j@medQ0%eKjbh8pO.^'LcCD,L*6P)3#odh%+r"J\$n:)LVlTrTIOm/oL'r
	#E&ce=k6Fiu8aXm1/:;hm?p#L^qI6J;D8?ZBMB_D14&ANkg9GMLR*Xs"/?@4VWUdJ,1MBB0[bECn33
	KZ1__A<"/u9(o<Sf@<$^stNom5GmA@5SIJ$^\D=(p8G;&!HNh)6lgYLfW6>#jE3aT_'W?L>Xr73'A#
	m<7:2<I)2%%Jk$'i-7@>Ml+rPk4?-&B7ph6*MjH9&DV+Uo=D(4f6)f(Z9"SdCXSlj^V?0?8][X1#pG
	[0-Dbk^rVg[Rc^PBH/8U_8QFCWdi&3#DT?k^_gU>S_]F^9g'7.>5F'hcYV%X?$[g4KPRF0=NW^$Z(L
	G'1aoAKLpqS!ei0kB29iHZJgJ(_`LbUX%/C@J6!+"mVIs6V)A,gbdt$K*g_X+Um(2\?XI=m'*tR%i"
	,kQIh51]UETI+HBA)DV:t/sl4<N*#^^=N.<B%00/$P>lNVlic"'Jc$p^ou^SLA\BS?`$Jla/$38;!#
	Q+K;T6T_?\3*d?$+27Ri'PYY-u]-gEMR^<d.ElNUY$#A@tX-ULd\IU&bfX]^T)a;<u7HgR!i2]GBpt
	SiZ1;JHl$jf3!k*jJlX$(ZroR:&!&8D[<-`g,)N96+6gSFVi$$Gr%h:1ioG%bZgmgbcp;2_&rF/[l"
	Qr^V@O-"j&UsEk)HgI'9`W31Wh"3^O,KrI/W'chm_@T!!^1"Y*Hknod`FiW&N@PIluYQdKILa)RK=W
	Ub4(Y)ao_5hG\K-+^73&AnBNQ+'D,6!KY/`F@6)`,V<qS#*-t?,F98]@h"8Y7Kj.%``Q=h4.L(m=Nd
	,%6Vs`ptRkJNBdbpk]$\>hR4"[5SF8$^:q=W([+TB`,%?4h7'ET[Y6F!KJ3fH"9BpILuUI#GoI.rl(
	_DAn[OiS_GkcL7QT`\p%Sos;F.W#_'g]3!!!!j78?7R6=>B
	ASCII85End
End
